#!/usr/bin/env python
#
#   __init__.py
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""Run all tests."""

import unittest

import resmantest
import timertest
import vectortest

__author__ = 'Ross Light'
__date__ = 'July 26, 2006'
__all__ = ['resmantest',
           'timertest',
           'vectortest',
           'test_suite',]

test_suite = unittest.TestSuite([resmantest.test_suite,
                                 timertest.test_suite,
                                 vectortest.test_suite,])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#!/usr/bin/env python
#
#   ui.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Game user interface based on widget objects

There are three types of coordinate systems when dealing with widgets:

Screen coordinates
    Absolute coordinates

Local coordinates
    Coordinates relative to the origin of the containing widget.  If the
    widget has no parent, then this is the same as screen coordinates.

Bounds coordinates
    Coordinates relative to the origin of the widget.  If the origin of the
    widget is (0, 0), then this is the same as local coordinates.

:Variables:
    left_align : int
        Left alignment constant for `TextWidget`
    center_align : int
        Center alignment constant for `TextWidget`
    right_align : int
        Right alignment constant for `TextWidget`
"""

import weakref

import pygame
from pygame.locals import *

from pymage import sprites

__author__ = 'Ross Light'
__date__ = 'July 20, 2006'
__all__ = ['left_align',
           'center_align',
           'right_align',
           'Widget',
           'Container',
           'TextWidget',
           'ImageWidget',
           'CursorWidget',
           'ButtonState',
           'StretchState',
           'Button',
           'PushButton',]
__docformat__ = 'reStructuredText'

# Constants

left_align = 0
center_align = 1
right_align = 2

class Widget(object):
    """
    Base class for all UI components.
    
    :IVariables:
        rect : ``pygame.Rect``
            The widget's rectangle in local coordinates.
        parent : `Widget`
            The parent widget.
    """
    def __init__(self,
                 parent=None,
                 rect=Rect(0, 0, 0, 0),
                 cache=False):
        """
        Initialize the widget.
        
        :Parameters:
            parent : `Widget`
                Parent widget
            rect : ``pygame.Rect``
                Initial button size and position
            cache : bool
                Whether to cache the button
        """
        self.__oldRect = None
        self.rect = rect
        self.__children = set()
        self.parent = parent
        self.__updates = [Rect(0, 0, self.rect.width, self.rect.height)]
        if cache:
            self.createCache()
        else:
            self.destroyCache()
        self.__active = False
    
    # Caching
    
    def createCache(self):
        """(Re)Creates a cache for the widget."""
        self.__cache = pygame.Surface(self.rect.size).convert_alpha()
        self.refresh()
    
    def destroyCache(self):
        """Destroys the widget's cache."""
        self.__cache = None
        self.refresh()
    
    def hasCache(self):
        """Returns whether the widget has a cache."""
        return bool(self.__cache is not None)
    
    # Child management
    
    def __len__(self):
        return len(self.__children)
    
    def __iter__(self):
        return iter(self.__children)
    
    def children(self):
        """
        Returns a list of all children.
        
        :ReturnType: list
        """
        return list(self.__children)
    
    def childTree(self, include_self=False, topdown=True):
        """
        Returns an iterator descending the entire child tree.
        
        :Keywords:
            include_self : bool
                Specifies whether the widget is included in the iterator.
            topdown : bool
                Specifies whether to go from widget to bottom or from bottom to
                widget.
        :ReturnType: iterator
        """
        if include_self and topdown:
            yield self
        for child in self.__children:
            tree = child.childTree(include_self=True,
                                   topdown=topdown)
            for subchild in tree:
                yield subchild
        if include_self and not topdown:
            yield self
    
    def addChild(self, child):
        """
        Adds a child widget.
        
        :Parameters:
            child : `Widget`
                New child to add
        """
        self.__children.add(child)
        child.parent = self
    
    def addChildren(self, children):
        """
        Adds children widgets.
        
        :Parameters:
            child : list of `Widget` objects
                New children to add
        """
        self.__children.update(frozenset(children))
        for child in children:
            child.parent = self
    
    def removeChild(self, child):
        """
        Removes a child widget.
        
        :Parameters:
            child : `Widget`
                Child to remove
        """
        self.__children.remove(child)
        del child.parent
    
    def removeAllChildren(self):
        """Removes all child widgets."""
        children = list(self.__children)
        self.__children.clear()
        for child in children:
            del child.parent
    
    def _getParent(self):
        """Retrieves the widget's parent."""
        if self.__parent is None:
            return None
        else:
            parent = self.__parent()
            if parent is None:
                self.__parent = None
            return parent
    
    def _setParent(self, new_parent):
        """Changes the widget's parent, and removes it from the old one."""
        # Remove self from old parent
        self._delParent()
        # Add self to new parent
        if new_parent is None:
            self.__parent = new_parent
        else:
            if self not in new_parent.children():
                new_parent.addChild(self)
            self.__parent = weakref.ref(new_parent)
    
    def _delParent(self):
        """Remove the widget from its parent."""
        try:
            if self.__parent is not None:
                parent = self.__parent()
                if parent is not None and self in parent:
                    parent.removeChild(self)
        except AttributeError:
            pass
        self.__parent = None
    
    @property
    def root(self):
        """The root of the hierarchy."""
        widget = self
        while True:
            parent = widget.parent
            if parent is None:
                break
            widget = parent
        return widget
    
    parent = property(_getParent, _setParent, _delParent)
    
    # Event management
    
    def handle(self, event):
        """
        Handles a single event.
        
        Override to perform custom event-handling.  Default implementation
        returns ``False``.
        
        :Parameters:
            event : ``pygame.event.Event``
                The event to handle.
        :Returns: ``True`` if the event was handled successfully, ``False``
                  otherwise.
        :ReturnType: bool
        """
        return False
    
    def processEvent(self, event):
        """
        Process an event.
        
        The event will start at a certain widget and traverse up the child
        hierarchy until the event is handled or the root is reached.  The
        starting widget is dependent on what the event type is:
        
        * If the event is a mouse event, the widget that the mouse is/was over.
        * Otherwise, the active widget.
        
        .. Note:: Do not override this method to perform event handling;
                  instead, override the `handle` method.
        
        :Parameters:
            event : ``pygame.event.Event``
                The event to process.
        :Returns: Whether the event was processed successfully.
        :ReturnType: bool
        """
        if event.type in (MOUSEBUTTONDOWN, MOUSEBUTTONUP, MOUSEMOTION):
            tree = self.root.childTree(include_self=True,
                                       topdown=False)
            for child in tree:
                if child.screenRect().collidepoint(*event.pos):
                    widget = child
                    break
            else:
                widget = None
        else:
            widget = self.root.activeWidget()
        while widget is not None:
            if widget.handle(event):
                return True
            widget = widget.parent
        else:
            return False
    
    def canFocus(self):
        """
        Returns whether the widget can be focused.
        
        Default returns ``False``.
        
        :ReturnType: bool
        """
        return False
    
    def focus(self, force_blur=False, force_focus=False):
        """
        Makes the widget the active one.
        
        .. Tip:: You may extend this implementation to perform custom handling,
                 but always call this implementation to actually focus the
                 widget.
        
        :Keywords:
            force_blur : bool
                Whether to force the currently active widget to blur.
            force_focus : bool
                Whether to force the widget to focus.
        :Returns: Whether the focus was successful.
        :ReturnType: bool
        """
        if force_focus or self.canFocus():
            activeWidget = self.root.activeWidget()
            if activeWidget is None or activeWidget.blur(force=force_blur):
                self.__active = True
                return True
            else:
                return False
        else:
            return False
    
    def canBlur(self):
        """
        Returns whether the widget can be blurred.
        
        Default returns ``True``.
        
        :ReturnType: bool
        """
        return True
    
    def blur(self, force=False):
        """
        Makes the widget inactive.
        
        .. Tip:: You may extend this implementation to perform custom handling,
                 but always call this implementation to actually blur the
                 widget.
        
        :Keywords:
            force : bool
                Whether to force the widget to blur.
        :Returns: Whether the blur was successful.
        :ReturnType: bool
        """
        if force or self.canBlur():
            self.__active = False
            return True
        else:
            return False
    
    def activeWidget(self):
        """
        Find the current active widget.
        
        .. Caution:: This only traverses from the receiver down.
        
        :Returns: The current active widget, or ``None`` if there is not one.
        :ReturnType: `Widget`
        """
        for child in self.childTree(include_self=True):
            if child.isActive():
                return child
        else:
            return None
    
    def isActive(self):
        """
        Returns whether the widget is active.
        
        :ReturnType: bool
        """
        return self.__active
    
    # Displaying
    
    def display(self, surface=None, origin=(0, 0)):
        """
        Display the widget on the given surface.
        
        .. Note:: You *shouldn't* have to override this method.  Instead, you
                  should override the `draw` method.
        
        :Parameters:
            surface : ``pygame.Surface``
                The surface to display on.  If this is not specified, the
                current display surface is used.
            origin : tuple
                The point in local space to use as the origin.  If not given,
                (0, 0) is used.
        :Returns: A list of ``pygame.Rect`` objects that need to be
                  updated
        :ReturnType: list of ``pygame.Rect`` objects
        """
        screenUpdates = []
        if surface is None:
            surface = pygame.display.get_surface()
        # Add old bounds to screen updates
        if self.__oldRect != self.rect:
            if self.__oldRect is not None:
                screenUpdates.append(self.__oldRect.move(origin))
            self.rectChanged(self.__oldRect, Rect(self.rect))
            if self.hasCache():
                self.createCache()
            screenUpdates.append(self.rect.move(origin))
            self.__oldRect = Rect(self.rect)
        # Calculate dirty region
        if self.__updates:
            region = self.__updates[0].unionall(self.__updates[1:])
            del self.__updates[:]
        else:
            region = None
        # Draw self
        if self.hasCache():
            # Using cache
            drawSurf = self.__cache
            if region is not None:
                self.draw(drawSurf, region)
        else:
            # Not using cache
            drawSurf = pygame.Surface(self.rect.size).convert_alpha(surface)
            self.draw(drawSurf, self.bounds())
        surface.blit(drawSurf, self.rect.move(origin))
        # Determine screen updates
        if region is not None:
            screenUpdates.append(self.btlrect(region, origin))
        # Display children
        childOrigin = self.rect.move(origin).topleft
        for child in self.__children:
            screenUpdates.extend(child.display(surface, childOrigin))
        return screenUpdates
    
    def draw(self, surface, rect):
        """
        Performs the actual drawing.
        
        Default implementation clears the surface to a transparent color.
        
        :Parameters:
            surface : ``pygame.Surface``
                Surface to draw onto.
            rect : ``pygame.Rect``
                Portion of the widget that needs to be refreshed.  This is only
                a guideline; you can update wherever you want.
        """
        surface.fill((0, 0, 0, 0), rect)
    
    def refresh(self, rect=None):
        """
        Inform the widget that a given portion needs to be refreshed.
        
        :Parameters:
            rect : ``pygame.Rect``
                The portion of the widget, in local space, that needs to be
                refreshed.  If not given, the entire widget will be refreshed.
        """
        if rect is None:
            rect = Rect(0, 0, self.rect.width, self.rect.height)
        self.__updates.append(Rect(rect))
    
    # Updating
    
    def update(self):
        """
        Updates the widget.
        
        Default implementation calls `update` on child widgets.
        """
        for child in self.__children:
            child.update()
    
    # Coordinates
    
    def stlrect(self, rect, origin=(0, 0)):
        """
        Converts from a screen rectangle to a local rectangle.
        
        :Parameters:
            rect : ``pygame.Rect``
                Rectangle to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted rectangle
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(rect)
        origin = self._localOrigin(origin)
        return rect.move(-origin[0], -origin[1])
    
    def stbrect(self, rect, origin=(0, 0)):
        """
        Converts from a screen rectangle to a bounds rectangle.
        
        :Parameters:
            rect : ``pygame.Rect``
                Rectangle to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted rectangle
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(rect)
        origin = self._boundsOrigin(origin)
        return rect.move(-origin[0], -origin[1])
    
    def ltsrect(self, rect, origin=(0, 0)):
        """
        Converts from a local rectangle to a screen rectangle.
        
        :Parameters:
            rect : ``pygame.Rect``
                Rectangle to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted rectangle
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(rect)
        origin = self._localOrigin(origin)
        return rect.move(origin[0], origin[1])
    
    def ltbrect(self, rect, origin=(0, 0)):
        """
        Converts from a local rectangle to a bounds rectangle.
        
        :Parameters:
            rect : ``pygame.Rect``
                Rectangle to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted rectangle
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(rect)
        origin = self.rect.topleft
        return rect.move(-origin[0], -origin[1])
    
    def btsrect(self, rect, origin=(0, 0)):
        """
        Converts from a bounds rectangle to a screen rectangle.
        
        :Parameters:
            rect : ``pygame.Rect``
                Rectangle to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted rectangle
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(rect)
        origin = self._boundsOrigin(origin)
        return rect.move(origin[0], origin[1])
    
    def btlrect(self, rect, origin=(0, 0)):
        """
        Converts from a bounds rectangle to a screen rectangle.
        
        :Parameters:
            rect : ``pygame.Rect``
                Rectangle to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted rectangle
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(rect)
        origin = self.rect.topleft
        return rect.move(origin[0], origin[1])
    
    def stlpoint(self, point, origin=(0, 0)):
        """
        Converts from a screen point to a local point.
        
        :Parameters:
            point : tuple
                Point to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted point
        :ReturnType: tuple
        """
        origin = self._localOrigin(origin)
        return (point[0] - origin[0], point[1] - origin[1])
    
    def stbpoint(self, point, origin=(0, 0)):
        """
        Converts from a screen point to a bounds point.
        
        :Parameters:
            point : tuple
                Point to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted point
        :ReturnType: tuple
        """
        origin = self._boundsOrigin(origin)
        return (point[0] - origin[0], point[1] - origin[1])
    
    def ltspoint(self, point, origin=(0, 0)):
        """
        Converts from a local point to a screen point.
        
        :Parameters:
            point : tuple
                Point to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted point
        :ReturnType: tuple
        """
        origin = self._localOrigin(origin)
        return (point[0] + origin[0], point[1] + origin[1])
    
    def ltbpoint(self, point, origin=(0, 0)):
        """
        Converts from a local point to a bounds point.
        
        :Parameters:
            point : tuple
                Point to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted point
        :ReturnType: tuple
        """
        origin = self.rect.topleft
        return (point[0] - origin[0], point[1] - origin[1])
    
    def btspoint(self, point, origin=(0, 0)):
        """
        Converts from a bounds point to a screen point.
        
        :Parameters:
            point : tuple
                Point to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted point
        :ReturnType: tuple
        """
        origin = self._boundsOrigin(origin)
        return (point[0] + origin[0], point[1] + origin[1])
    
    def btlpoint(self, point, origin=(0, 0)):
        """
        Converts from a bounds point to a local point.
        
        :Parameters:
            point : tuple
                Point to convert
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The converted point
        :ReturnType: tuple
        """
        origin = self.rect.topleft
        return (point[0] + origin[0], point[1] + origin[1])
    
    def _localOrigin(self, origin=(0, 0)):
        """Returns the local origin as screen coordinates."""
        origin = self.screenRect(origin).topleft
        return (origin[0] - self.rect.x, origin[1] - self.rect.y)
    
    def _boundsOrigin(self, origin=(0, 0)):
        """Returns the bounds origin as screen coordinates."""
        return self.screenRect(origin).topleft
    
    def screenRect(self, origin=(0, 0)):
        """
        Calculates the screen rectangle.
        
        :Parameters:
            origin : tuple
                The point to consider the origin; default is (0, 0).
        :Returns: The widget's rectangle in screen coordinates
        :ReturnType: ``pygame.Rect``
        """
        rect = Rect(self.rect)
        parent = self.parent
        while parent is not None:
            rect.move_ip(parent.rect.topleft)
            parent = parent.parent
        rect.move_ip(origin)
        return rect
    
    def bounds(self):
        """
        Returns the rectangle that can be drawn in.
        
        :ReturnType: ``pygame.Rect``
        """
        return Rect(0, 0, self.rect.width, self.rect.height)
    
    # Dynamic sizing
    
    def optimalSize(self):
        """
        Returns the optimal size of the widget.
        
        Default returns ``None``, but subclasses can override to return an
        actual value.
        
        :Returns: The optimal size, or ``None`` if one cannot be calculated.
        :ReturnType: tuple
        """
        return None
    
    def pack(self, anchor=None):
        """
        Changes the widget's rect to the optimal size.
        
        If `optimalSize` returns None, this does nothing.
        
        .. Tip:: For best results, specify a corner (e.g. topleft) for anchor.
        
        :Parameters:
            anchor : string
                A corner/edge attribute name that will remain constant after the
                resize.
        """
        if anchor is not None:
            anchorValue = getattr(self.rect, anchor)
        size = self.optimalSize()
        if size is not None:
            self.rect.size = size
            if anchor is not None:
                setattr(self.rect, anchor, anchorValue)
    
    def rectChanged(self, old_rect, new_rect):
        """
        Hook method for the rectangle changing.
        
        This is called before `display` does any drawing.
        
        :Parameters:
            old_rect : ``pygame.Rect``
                The old rectangle in local coordinates
            new_rect : ``pygame.Rect``
                The new rectangle in local coordinates
        """
        pass
    
    def _getRect(self):
        return self.__rect
    
    def _setRect(self, new_rect):
        self.__rect = Rect(new_rect)
    
    rect = property(_getRect, _setRect)

class Container(Widget):
    """
    A widget that is designed to contain other widgets.
    
    :IVariables:
        border : int
            The border width (in pixels).
        borderColor : tuple
            The border color as an RGBA 0-255 tuple
        bgColor : tuple
            The background color as an RGBA 0-255 tuple
    """
    def __init__(self,
                 border=0,
                 border_color=(0, 0, 0, 255),
                 bg_color=(0, 0, 0, 0),
                 *args, **kw):
        super(Container, self).__init__(*args, **kw)
        self.border = border
        self.borderColor = border_color
        self.bgColor = bg_color
    
    def draw(self, surface, rect):
        surface.fill(self.bgColor, rect)
        if self.border > 0:
            pygame.draw.rect(surface,
                             self.borderColor,
                             self.bounds(),
                             self.border)
    
    def optimalSize(self):
        childRects = [child.rect for child in self.children()]
        if childRects:
            allChildRect = childRects[0].union_all(childRects[1:])
            return (allChildRect.x + allChildRect.width,
                    allChildRect.y + allChildRect.height)
        else:
            return None

class TextWidget(Widget):
    """
    A widget that can display a single line of text.
    
    :IVariables:
        text : string
            The text the widget displays.
        font : ``pygame.font.Font``
            The font the text is displayed as.  Default is the pygame default
            font at size 14.
        fgColor : tuple
            The text color as an RGBA 0-255 tuple.  Default is black.
        bgColor : tuple
            The background color as an RGBA 0-255 tuple.  Default is clear.
        antialias : bool
            Whether or not to antialias the text.  Default is ``True``.
        align : int
            The alignment as a constant (i.e. `left_align`, `center_align`, or
            `right_align`).  The default is left alignment.
        shadow : bool
            Whether or not to add a *cheap* shadow.  Default is ``False``.
        shadowColor : tuple
            The shadow color as an RGBA 0-255 tuple.  Default is a
            half-transparent black.
    """
    fgColor = (0, 0, 0, 255)
    bgColor = (0, 0, 0, 0)
    antialias = True
    shadow = False
    shadowColor = (0, 0, 0, 128)
    
    def __init__(self, text='', font=None, *args, **kw):
        """
        Initializes a text widget.
        
        The additional variables can be specified via keyword arguments.
        """
        # Get keyword options
        self.fgColor = kw.pop('fgColor', self.fgColor)
        self.bgColor = kw.pop('bgColor', self.bgColor)
        self.antialias = kw.pop('antialias', self.antialias)
        self.align = kw.pop('align', left_align)
        self.shadow = kw.pop('shadow', self.shadow)
        # Perform superclass initialization
        super(TextWidget, self).__init__(*args, **kw)
        # Perform initialization
        self.text = text
        if font is None:
            font = pygame.font.Font(None, 14)
        self.font = font
    
    def draw(self, surface, rect):
        # Render text
        surface.fill(self.bgColor, rect)
        rendered = self.font.render(self.text, self.antialias, self.fgColor)
        # Determine render position
        textRect = rendered.get_rect()
        alignPoints = {left_align: 'topleft',
                       center_align: 'midtop',
                       right_align: 'topright',}
        setattr(textRect, alignPoints[self.align],
                getattr(self.bounds(), alignPoints[self.align]))
        # Render shadow
        if self.shadow:
            renderedShadow = self.font.render(self.text,
                                              self.antialias,
                                              self.shadowColor)
            shadowRect = textRect.move(0, 2)
        else:
            renderedShadow = None
            shadowRect = None
        # Blit text to screen, if necessary
        if renderedShadow is not None and shadowRect.colliderect(rect):
            surface.blit(renderedShadow, shadowRect)
        if textRect.colliderect(rect):
            surface.blit(rendered, textRect)
    
    def optimalSize(self):
        return self.font.size(self.text)
    
    def _getText(self):
        """Retrieve text value."""
        return self._text
    
    def _setText(self, text):
        """Change text value and refresh."""
        self._text = text
        self.refresh()
    
    def _getFont(self):
        """Retrieve font."""
        return self._font
    
    def _setFont(self, font):
        """Change font and refresh."""
        self._font = font
        self.refresh()
    
    def _getAlign(self):
        """Retrieve alignment."""
        return self._align
    
    def _setAlign(self, align):
        """Change alignment and validate."""
        if align in (left_align, center_align, right_align):
            self._align = align
        else:
            raise ValueError("Alignment %r is not valid" % align)
    
    text = property(_getText, _setText)
    font = property(_getFont, _setFont)
    align = property(_getAlign, _setAlign)

class ImageWidget(Widget):
    """
    A widget that displays an image.
    
    :IVariables:
        image : ``pygame.Surface``
            The image the widget displays.
        stretch : bool
            Whether to stretch the image to fit the size.
    """
    def __init__(self, image, stretch=False, *args, **kw):
        super(ImageWidget, self).__init__(*args, **kw)
        self.image = image
        self.stretch = stretch
    
    def draw(self, surface, rect):
        super(ImageWidget, self).draw(surface, rect)
        image = self.image
        if self.stretch and self.rect.size != image.get_size():
            image = pygame.transform.scale(self.image, rect.size)
        surface.blit(image, (0, 0), rect)
    
    def getImage(self):
        """
        Retrieves the current image.
        
        You can also use the ``image`` property.
        """
        return self.__image
    
    def setImage(self, image, tryIM=True):
        """
        Changes the current image.
        
        You can also use the ``image`` property, but it always tries
        `sprites.ImageManager`.
        """
        self.__image = sprites.getImage(image, tryIM)
        self.refresh()
    
    def optimalSize(self):
        return self.image.get_size()
    
    image = property(getImage, setImage)

class CursorWidget(ImageWidget):
    """
    A widget that displays an image, following the cursor.
    
    :IVariables:
        hotspot : tuple
            The location of the cursor's "hotspot".  When the mouse is clicked,
            the hotspot is the place where the mouse position is registered.
    """
    def __init__(self, hotspot=(0, 0), *args, **kw):
        super(CursorWidget, self).__init__(*args, **kw)
        self.hotspot = hotspot
        self.followMouse()
    
    def handle(self, event):
        if event.type == MOUSEMOTION:
            self.followMouse(event.pos)
            return True
        else:
            return False
    
    def followMouse(self, pos=None):
        """
        Called to follow the mouse.
        
        :Parameters:
            pos : tuple
                The position where the mouse is.  If not given, the position is
                retrieved with ``pygame.mouse.get_pos``.
        """
        if pos is None:
            pos = pygame.mouse.get_pos()
        self.rect.topleft = self.stlpoint((pos[0] - self.hotspot[0],
                                           pos[1] - self.hotspot[1]))

class ButtonState(object):
    """A button state."""
    def setState(self, button):
        """
        Called to change the button into the given state.
        
        :Parameters:
            button : `Button`
                The button whose state is changing.
        """
        pass
    
    def unsetState(self, button):
        """
        Called to undo the effects of `setState`.
        
        :Parameters:
            button : `Button`
                The button whose state is being undone.
        """
        pass
    
    def optimalSize(self):
        """Returns the optimal size of the button, or ``None``."""
        return None

class StretchState(ButtonState):
    """
    A button state in which three images are used.
    
    Two are used for caps, and the center is stretched to fit the text.
    
    :IVariables:
        text : string
            The text being displayed on the button.
        textColor : tuple
            The text color as an RGBA 0-255 tuple.
        imgL : ``pygame.Surface``
            The left image cap.
        imgC : ``pygame.Surface``
            The stretched center image.
        imgR : ``pygame.Surface``
            The right image cap.
    """
    def __init__(self, text, textColor, imgL, imgC, imgR):
        self.text = text
        self.textColor = textColor
        self.imgL = sprites.getImage(imgL)
        self.imgC = sprites.getImage(imgC)
        self.imgR = sprites.getImage(imgR)
    
    def setState(self, button):
        # Calculate left cap rect
        lrect = Rect(0, 0, 0, 0)
        lrect.size = self.imgL.get_size()
        lrect.topleft = (0, 0)
        # Calculate right cap rect
        rrect = Rect(0, 0, 0, 0)
        rrect.size = self.imgR.get_size()
        rrect.topright = (button.rect.width, 0)
        # Calculate center cap rect
        crect = Rect(0, 0, 0, 0)
        crect.height = self.imgC.get_height()
        crect.width = rrect.left - lrect.right
        crect.topleft = lrect.topright
        # Add children
        ImageWidget(self.imgL,
                    parent=button,
                    rect=lrect)
        ImageWidget(self.imgC,
                    stretch=True,
                    parent=button,
                    rect=crect)
        ImageWidget(self.imgR,
                    parent=button,
                    rect=rrect)
        TextWidget(self.text,
                   font=self.getFont(),
                   align=center_align,
                   fgColor=self.textColor,
                   parent=button,
                   rect=crect)
    
    def unsetState(self, button):
        button.removeAllChildren()
        
    def getFont(self):
        return pygame.font.Font(None, self.imgC.get_height())
    
    def optimalSize(self):
        font = self.getFont()
        width = (self.imgL.get_width() +
                 font.size(self.text)[0] +
                 self.imgR.get_width())
        height = self.imgC.get_height()
        return (width, height)

class Button(Widget):
    """
    Abstract superclass for buttons.
    
    :IVariables:
        state : `ButtonState`
            The button's current state
        callback : function
            The button's action callback
        callArgs : tuple
            Additional arguments passed to the button's callback
        callKw : dict
            Additional keyword arguments passed to the button's callback
        nav_up : `Widget`
            The widget activated by pressing the up arrow key.
        nav_down : `Widget`
            The widget activated by pressing the down arrow key.
        nav_left : `Widget`
            The widget activated by pressing the left arrow key.
        nav_right : `Widget`
            The widget activated by pressing the right arrow key.
    """
    def __init__(self,
                 state,
                 callback=None,
                 args=[],
                 kw={},
                 *initArgs, **initKw):
        """
        Initializes the button.
        
        :Parameters:
            state : `ButtonState`
                Initial button state.
            callback : function
                Button's action callback
            args : tuple
                Additional callback arguments
            kw : dict
                Additional callback keyword arguments
            parent : `Widget`
                Parent widget
            rect : ``pygame.Rect``
                Initial button size and position
            cache : bool
                Whether to cache the button
        :Keywords:
            up : `Widget`
                The widget activated by pressing the up arrow key.
            down : `Widget`
                The widget activated by pressing the down arrow key.
            left : `Widget`
                The widget activated by pressing the left arrow key.
            right : `Widget`
                The widget activated by pressing the right arrow key.
        """
        self.nav_up = initKw.pop('up', None)
        self.nav_down = initKw.pop('down', None)
        self.nav_left = initKw.pop('left', None)
        self.nav_right = initKw.pop('right', None)
        super(Button, self).__init__(*initArgs, **initKw)
        self.setState(state)
        self.callback = callback
        self.callArgs = args
        self.callKw = kw
    
    def handle(self, event):
        keyButtons = {K_UP: self.nav_up,
                      K_DOWN: self.nav_down,
                      K_LEFT: self.nav_left,
                      K_RIGHT: self.nav_right,}
        hatButtons = {(0, 1): self.nav_up,
                      (0, -1): self.nav_down,
                      (-1, 0): self.nav_left,
                      (1, 0): self.nav_right,}
        if event.type == KEYUP and keyButtons.get(event.key) is not None:
            keyButtons[event.key].focus()
            return True
        elif event.type == JOYHATMOTION and \
             hatButtons.get(event.value) is not None:
            hatButtons[event.value].focus()
            return True
        else:
            return False
    
    def setState(self, new_state):
        try:
            oldState = self.state
        except AttributeError:
            pass
        else:
            if oldState == new_state:
                return
            oldState.unsetState(self)
        self.state = new_state
        self.state.setState(self)
    
    def perform(self):
        """Calls the button's callback."""
        if self.callback is not None:
            self.callback(self, *self.callArgs, **self.callKw)
    
    def optimalSize(self):
        """Sizes the button to the optimum size."""
        return self.state.optimalSize()
    
    def rectChanged(self, old_rect, new_rect):
        self.state.unsetState(self)
        self.state.setState(self)
    
    def __getattribute__(self, name):
        value = super(Button, self).__getattribute__(name)
        if name.startswith('nav_') and value is not None:
            value = value()
            if value is None:
                setattr(self, name, None)
        return value
    
    def __setattr__(self, name, value):
        if name.startswith('nav_') and value is not None:
            value = weakref.ref(value)
        super(Button, self).__setattr__(name, value)

class PushButton(Button):
    """
    Button activated by clicking or pressing Enter while active.
    
    :IVariables:
        normState : `ButtonState`
            The inactive state
        rollState : `ButtonState`
            The active state
    """
    def __init__(self,
                 normState,
                 rollState=None,
                 *args, **kw):
        """
        Initializes the button.
        
        :Parameters:
            normState : `ButtonState`
                The inactive state
            rollState : `ButtonState`
                The active state
            callback : function
                Button's action callback
            args : tuple
                Additional callback arguments
            kw : dict
                Additional callback keyword arguments
            parent : `Widget`
                Parent widget
            rect : ``pygame.Rect``
                Initial button size and position
            cache : bool
                Whether to cache the button
        :Keywords:
            up : `Widget`
                The widget activated by pressing the up arrow key.
            down : `Widget`
                The widget activated by pressing the down arrow key.
            left : `Widget`
                The widget activated by pressing the left arrow key.
            right : `Widget`
                The widget activated by pressing the right arrow key.
        """
        super(PushButton, self).__init__(normState, *args, **kw)
        self.normState = normState
        if rollState is None:
            self.rollState = self.normState
        else:
            self.rollState = rollState
        self._lastMousePos = pygame.mouse.get_pos()
    
    def handle(self, event):
        if super(PushButton, self).handle(event):
            return True
        elif (event.type == MOUSEBUTTONUP and event.button == 1) or \
             (event.type == KEYUP and event.key == K_RETURN) or \
             (event.type == JOYBUTTONUP and event.button == 0):
            self.perform()
            return True
        else:
            return False
    
    def update(self):
        mousePos = pygame.mouse.get_pos()
        if mousePos != self._lastMousePos:
            self._lastMousePos = mousePos
            mouseInButton = self.screenRect().collidepoint(mousePos)
            if not self.isActive() and mouseInButton:
                self.focus()
            elif self.isActive() and not mouseInButton:
                self.blur()
    
    def canFocus(self):
        return True
    
    def focus(self, *args, **kw):
        success = super(PushButton, self).focus(*args, **kw)
        if success:
            self.setState(self.rollState)
        return success
    
    def blur(self, *args, **kw):
        success = super(PushButton, self).blur(*args, **kw)
        if success:
            self.setState(self.normState)
        return success



#!/usr/bin/env python
#
#   sprites.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Defines game sprites

:Variables:
    im : `ImageManager`
        Global image manager
"""

import warnings

import pygame
from pygame.locals import *

from pymage import resman

__author__ = 'Ross Light'
__date__ = 'May 22, 2006'
__all__ = ['getImage',
           'ImageManager',
           'im',
           'Sprite',
           'Animation',]
__docformat__ = 'reStructuredText'

def getImage(image, tryIM=True):
    """
    Retrieves an image.
    
    Provides a standard way to either pass an image or a string to a function.
    
    :Parameters:
        image : string
            Name of the image resource
        tryIM : bool
            Whether to use the image manager.  If ``False`` or the image manager
            cannot find it, the image is assumed to be a path.
    :Returns: The image requested
    :ReturnType: ``pygame.Surface``
    """
    if isinstance(image, pygame.Surface):
        return image
    else:
        if tryIM:
            try:
                return im.load(image)
            except KeyError:
                pass
        return pygame.image.load(image)

class ImageManager(resman.Submanager):
    """Game image manager."""
    resourceType = resman.ImageResource

    def loadImage(self, *args, **kw):
        """
        Loads an image from disk, using a cached representation, if possible.
        
        .. Warning::
           `loadImage` method is deprecated, in favor of the new Submanager
           API.  Use `load` instead.
        
        :Parameters:
            key : string
                Name of the image resource
        :Returns: The image requested
        :ReturnType: ``pygame.Surface``
        """
        warnings.warn("loadImage is deprecated; use load.",
                      DeprecationWarning,
                      stacklevel=2)
        return self.load(*args, **kw)

im = ImageManager()

class Sprite(pygame.sprite.Sprite, object):
    """
    Abstract superclass for sprites.
    
    :CVariables:
        hpadding : int
            The horizontal padding for the collision box.  See `collideBox`.
        vpadding : int
            The vertical padding for the collision box.  See `collideBox`.
        angleTolerance : float
            The angle tolerance for using the initial image
        clamp : bool
            Whether to clamp the sprite to the screen boundaries
    :IVariables:
        image : ``pygame.Surface``
            The sprite's image
        rect : ``pygame.Rect``
            The sprite's position
        angle : float
            The sprite's angle (in counterclockwise degrees)
        area : ``pygame.Rect``
            The clamping area
    """
    hpadding = vpadding = 0
    angleTolerance = 0.5
    clamp = True
    
    def __init__(self, image=None):
        """
        Initializes a sprite.
        
        :Parameters:
            image : string
                The initial image to use
        """
        pygame.sprite.Sprite.__init__(self)
        if image is None:
            image = self.image
        self.setImage(image)
        self.rect = self.image.get_rect()
        self.area = pygame.display.get_surface().get_rect()
        self.angle = 0.0
    
    def setImage(self, image, tryIM=True):
        """
        Changes the current image and the revert image variables.
        
        :Parameters:
            image : string
                The image name
            tryIM : bool
                Whether to use the image manager
        """
        self.image = self._image = getImage(image, tryIM).convert_alpha()
    
    def collideBox(self):
        """
        Returns the box used for checking with the `touches` method.
        
        By default, this uses the `hpadding` and `vpadding` to construct an
        inset box.  Override to have a different collide box.
        
        :Returns: The collision box
        :ReturnType: ``pygame.Rect``
        """
        # Multiply by two to get all-around coverage
        return self.rect.inflate(self.hpadding * -2, self.vpadding * -2)
    
    def touches(self, other):
        """
        Returns whether the other sprite is actually touching the receiver.
        
        :Parameters:
            other : `Sprite`
                The sprite to test collision with
        :ReturnType: bool
        """
        return self.collideBox().colliderect(other.collideBox())
    
    def updateWithVector(self, vector, clamp=None):
        """
        Moves the sprite with the given vector.
        
        :Parameters:
            vector : `pymage.vector.Vector`
                The vector describing where to move
            clamp : bool
                Whether to clamp to `area`.  If not specified, this depends on
                the `clamp` attribute.
        """
        self.rect.x += vector.x
        self.rect.y += vector.y
        if clamp is None:
            clamp = self.clamp
        if clamp:
            self.rect = self.rect.clamp(self.area)
    
    def rotateImage(self):
        """Rotates the sprite's image to the proper angle."""
        if abs(self.angle) < self.angleTolerance:
            self.image = self._image
        else:
            self.image = pygame.transform.rotate(self._image, self.angle)
        self.rect.size = self.image.get_size()

class Animation(Sprite):
    """
    Superclass for ambient animations.
    
    :IVariables:
        frames : list of ``pygame.Surface``s or strings
            Individual frames of the animation
        loop : bool
            Whether the animation should continuously play or whether it should
            kill itself after one run
    """
    loop = False
    
    def __init__(self, frames=None, loop=None):
        """
        Initializes an animation.
        
        :Parameters:
            frames : list of ``pygame.Surface``s or strings
                The individual frames of the animation.  If not specified, the
                `frames` class variable is used.
            loop : bool
                Whether the animation should continuously play or whether it
                should kill itself after one run.  If not specified, it uses
                the `loop` class variable.
        """
        if frames is None:
            frames = self.frames
        self.frames = list(frames)
        if loop is not None:
            self.loop = loop
        self.frameNum = 0
        super(Animation, self).__init__(frames[0])
    
    def update(self):
        """
        Updates the sprite.
        
        Default implementation advances to the next frame.
        """
        self.advance()
    
    def advance(self):
        """Advances to the next frame and dies if it reaches the end."""
        self.frameNum += 1
        if self.frameNum >= len(self.frames):
            if self.loop:
                self.frameNum = 0
            else:
                self.kill()
        else:
            self.setImage(self.frames[self.frameNum])



#!/usr/bin/env python
#
#   __init__.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Various utilities to simplify Pygame_ development.

:Author: Ross Light
:Copyright: Copyright (C) 2006-2007 Ross Light
:Contact: rlight2@gmail.com
:License:
    pymage is free software; you can redistribute it and/or modify it under the
    terms of the `GNU Lesser General Public License`_ as published by the `Free
    Software Foundation`_; either version 3 of the License, or (at your option)
    any later version.
    
    pymage is distributed in the hope that it will be useful, but **WITHOUT ANY
    WARRANTY**; without even the implied warranty of **MERCHANTABILITY** or
    **FITNESS FOR A PARTICULAR PURPOSE**.  See the `GNU Lesser General Public
    License`_ for more details.
    
    You should have received a copy of the `GNU Lesser General Public License`_
    along with this library.  If not, see http://www.gnu.org/licenses/.

.. _Pygame: http://www.pygame.org/
.. _GNU Lesser General Public License: http://www.gnu.org/licenses/lgpl.html
.. _Free Software Foundation: http://fsf.org/
"""

__author__ = 'Ross Light'
__date__ = 'July 20, 2006'
__all__ = ['config',
           'joystick',
           'resman',
           'sound',
           'sprites',
           'states',
           'timer',
           'ui',
           'vector',]
__docformat__ = 'reStructuredText'
__version__ = '0.2.2'

from pymage import config
from pymage import joystick
from pymage import resman
from pymage import sound
from pymage import sprites
from pymage import states
from pymage import timer
from pymage import ui
from pymage import vector



#!/usr/bin/env python
#
#   states.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""Manages game states (e.g. paused, levels...)"""

from __future__ import division

import os
import sys

try:
    from OpenGL.GL import *
    from OpenGL.GLU import *
except ImportError:
    pass
import pygame
from pygame.locals import *

from pymage import sprites
from pymage import sound
from pymage import ui

__author__ = 'Ross Light'
__date__ = 'May 22, 2006'
__docformat__ = 'reStructuredText'
__all__ = ['State',
           'GLState',
           'Paused',
           'Menu',
           'Game',
           'FakeClock',
           'TwistedGame',]

class State(object):
    """
    Game state that handles events and displays on a surface.
    
    :CVariables:
        bgColor : tuple
            The state's clear color as an RGB 0-255 tuple
        quitOnEscape : bool
            Whether pressing escape quits the game
    """
    bgColor = (0, 0, 0)
    quitOnEscape = False
    
    def __init__(self, bgColor=None):
        if bgColor is not None:
            self.bgColor = bgColor
    
    def handle(self, event):
        """
        Hook method for handling events.
        
        .. Note:: This implementation *should* always be called.
        
        :Parameters:
            event : ``pygame.event.Event``
                Event to handle
        """
        if event.type == QUIT:
            sys.exit()
        elif self.quitOnEscape and \
             event.type == KEYDOWN and \
             event.key == K_ESCAPE:
            sys.exit()
    
    def update(self, game):
        """
        Hook method for updating the state.
        
        :Parameters:
            game : `Game`
                Current game
        """
        pass
    
    def firstDisplay(self, screen):
        """
        Hook method for displaying the `State` for the first time.
        
        By default, this merely fills the screen with the background color.
        
        :Parameters:
            screen : ``pygame.Surface``
                Surface to draw to
        """
        screen.fill(self.bgColor)
        pygame.display.flip()       # Swap buffers
    
    def display(self, screen):
        """
        Hook method for drawing the state.
        
        :Parameters:
            screen : ``pygame.Surface``
                Surface to draw to
        """
        pass

class GLState(State):
    """
    Game state with OpenGL-specific drawing.
    
    :CVariables:
        fovy : float
            Horizontal visible field-of-view angle
        clipNear : float
            Near clipping distance
        clipFar : float
            Far clipping distance
        bgColor : tuple
            Clear color as RGBA 0.0-1.0 tuple
    :IVariables:
        camPos : tuple
            Camera position as a (x, y, z) tuple
        camAim : tuple
            Camera aim as a (x, y, z) tuple
        camUpv : tuple
            Camera up vector as a (x, y, z) tuple
    """
    # Camera
    fovy = 60.0
    clipNear = 0.1
    clipFar = 100.0
    bgColor = (0.0, 0.0, 0.0, 1.0)
    
    def __init__(self, *args, **kw):
        super(GLState, self).__init__(*args, **kw)
        self.camPos = (2.0, 0.0, 0.0)
        self.camAim = (0.0, 0.0, 0.0)
        self.camUpv = (0.0, 5.0, 0.0)
    
    def initGL(self):
        """Called on first display to initialize OpenGL."""
        surf = pygame.display.get_surface()
        self.setupClearColor()
        self.setupCamera(surf.get_width(), surf.get_height())
    
    def setupGL(self):
        """Called on subsequent displays to setup OpenGL."""
        surf = pygame.display.get_surface()
        self.setupClearColor()
        self.setupCamera(surf.get_width(), surf.get_height())
    
    def setupClearColor(self):
        """Sets the OpenGL clear color from the `bgColor` attribute."""
        r, g, b = self.bgColor[:3]
        if len(self.bgColor) >= 4:
            a = self.bgColor[4]
        else:
            a = 1.0
        glClearColor(r, g, b, a)
    
    def setupCamera(self, width, height):
        """
        Sets up the OpenGL camera.
        
        This includes setting up the viewport, perspective, and camera position.
        
        :Parameters:
            width : int
                Width of viewport
            height : int
                Height of viewport
        """
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(self.fovy,
                       width / height,
                       self.clipNear,
                       self.clipFar)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        gluLookAt(self.camPos[0], self.camPos[1], self.camPos[2],
                  self.camAim[0], self.camAim[1], self.camAim[2],
                  self.camUpv[0], self.camUpv[1], self.camUpv[2])
    
    def firstDisplay(self, screen):
        """
        Called on first display.
        
        Calls `initGL`, does a ``glClear``, and flips the buffer.
        
        :Parameters:
            screen : ``pygame.Surface``
                Surface to draw to
        """
        self.initGL()
        glClear(GL_COLOR_BUFFER_BIT)
        pygame.display.flip()
    
    def display(self, screen):
        """
        Called on displaying.
        
        Calls `setupGL`, does a ``glClear``, and flips the buffer.
        
        :Parameters:
            screen : ``pygame.Surface``
                Surface to draw to
        """
        self.setupGL()
        glClear(GL_COLOR_BUFFER_BIT)
        pygame.display.flip()

class Paused(State):
    """
    Abstract superclass for simple pause screens that can show text and/or an
    image.
    
    :CVariables:
        image : ``pygame.Surface``
            Image to display in background
        useIM : bool
            Whether to use the image manager to get the image
        text : string
            String to display
        fontSize : int
            Size of the text font
        textColor : tuple
            Text color as an RGBA 0-255 tuple.
        excludedKeys : list
            Keys to ignore
    :IVariables:
        finished : bool
            Whether the state is finished
    """
    finished = False
    image = None
    useIM = True
    text = ""
    fontSize = 48
    textColor = (0, 0, 0)
    showCursor = False
    excludedKeys = []
    
    def __init__(self,
                 bg_color=None,
                 text=None,
                 image=None,
                 fontSize=None,
                 useIM=None):
        super(Paused, self).__init__(bg_color)
        # Set up text
        if text is not None:
            self.text = text
        if fontSize is not None:
            self.fontSize = fontSize
        # Set up image
        if image is not None:
            self.image = image
        if useIM is not None:
            self.useIM = useIM
        if self.useIM and self.image is not None:
            self.image = sprites.ImageManager.loadImage(self.image)
    
    def handle(self, event):
        """Stops the pause screen after a click or key press."""
        super(Paused, self).handle(event)
        if event.type in (MOUSEBUTTONUP, JOYBUTTONUP) or \
           event.type == KEYUP and event.key not in self.excludedKeys:
            self.finished = True
    
    def update(self, game):
        """Changes to the next state, if we're done."""
        if self.finished:
            pygame.mouse.set_visible(False)
            game.changeToState(self.nextState())
    
    def firstDisplay(self, screen):
        """Do actual display."""
        screen.fill(self.bgColor)   # Clear screen
        font = pygame.font.Font(None, self.fontSize)
        lines = self.text.strip().splitlines()
        # Calculate positioning
        height = len(lines) * font.get_linesize()
        center, top = screen.get_rect().center
        top -= height // 2
        # Show image
        if self.image:
            if isinstance(self.image, basestring):
                image = pygame.image.load(self.image).convert_alpha()
            else:
                image = self.image
            imageRect = image.get_rect()
            top += imageRect.height // 2
            imageRect.midbottom = center, top - 20
            screen.blit(image, imageRect)   # show the image
        # Render lines
        antialias = True
        for line in lines:
            text = font.render(line.strip(), antialias, self.textColor)
            textRect = text.get_rect()
            textRect.midtop = center, top
            screen.blit(text, textRect)
            top += font.get_linesize()
        # Show cursor (if necessary)
        pygame.mouse.set_visible(self.showCursor)
        pygame.display.flip()   # swap buffers
    
    def nextState(self):
        """
        Hook method for advancing to the next state.
        
        :Raises NotImplementedError: if not overridden
        :ReturnType: `State`
        """
        raise NotImplementedError("Override nextState")

class Menu(State):
    """
    Abstract superclass for menu screens.
    
    :IVariables:
        newState : `State`
            The next state to advance to
        container : `ui.Container`
            The root widget
        cursor : `ui.CursorWidget`
            Visible cursor
    """
    def __init__(self, *args, **kw):
        super(Menu, self).__init__(*args, **kw)
        initRect = pygame.display.get_surface().get_rect()
        self.newState = None
        self.container = ui.Container(bg_color=self.bgColor,
                                      rect=initRect)
        self.cursor = None
        self.body()
    
    def body(self):
        """Adds the menu's contents."""
        pass
    
    def handle(self, event):
        """Dispatches a UI event."""
        super(Menu, self).handle(event)
        self.container.processEvent(event)
        if self.cursor is not None:
            self.cursor.handle(event)
    
    def update(self, game):
        """
        Updates the UI widgets.
        
        If you change the `newState` attribute to a non-``None`` value, it will
        become the new state.
        """
        self.container.update()
        if self.cursor is not None:
            self.cursor.update()
        if self.newState is not None:
            game.changeToState(self.newState)
            self.newState = None
    
    def firstDisplay(self, screen):
        """Sets up the container."""
        self.container.refresh()
        self.container.rect.size = screen.get_size()
        super(Menu, self).firstDisplay(screen)
    
    def display(self, screen):
        """Redisplays any of the widgets that need it."""
        super(Menu, self).display(screen)
        updates = self.container.display(screen)
        if self.cursor is not None:
            updates += self.cursor.display(screen)
        pygame.display.update(updates)

class Game(object):
    """
    Game object that manages the state machine.
    
    :CVariables:
        screenSize : tuple
            Screen size (in pixels) as a (width, height) tuple
        ticks : int
            Capped number of frames per second
        flags : int
            Flags to pass to the display, as described in
            ``pygame.display.set_mode``.
    :IVariables:
        state : `State`
            Current state
        nextState : `State`
            State to advance to
        running : bool
            Whether the game is running
        clock : ``pygame.time.Clock``
            FPS timer
    """
    screenSize = (800, 600)
    ticks = 60
    flags = 0
    _globalGame = None
    
    @classmethod
    def getGame(cls):
        """
        Retrieves the current game.
        
        :ReturnType: `Game`
        """
        return cls._globalGame
    
    def __init__(self, root_dir, **kw):
        """
        Initialize game.
        
        :Parameters:
            root_dir : string
                Directory to chdir to.  You can even pass sys.argv[0].
        """
        # Change to root directory (for relative paths)
        os.chdir(os.path.abspath(os.path.dirname(root_dir)))
        # Start with no state
        self.state = self.nextState = None
        self.running = False
        self.screen = None
        self.clock = None
        # Set up instance variables
        self.__dict__.update(kw)
        # Set up global game
        if type(self)._globalGame is None:
            type(self)._globalGame = self
    
    def changeToState(self, state):
        """
        Changes to a state on the next event loop iteration.
        
        :Parameters:
            state : `State`
                State to change to
        """
        self.nextState = state
    
    def run(self):
        """
        Run main event loop.
        
        You *shouldn't* have to override this...
        """
        pygame.init()
        self.screen = pygame.display.set_mode(self.screenSize, self.flags)
        self.clock = pygame.time.Clock()
        self.preloop()
        self.running = True
        while self.running:
            try:
                self.iterate()
            except SystemExit:
                self.end()
        self.postloop()
    
    def end(self):
        """Ends the game."""
        self.running = False
    
    def preloop(self):
        """Hook method to do something before the event loop starts."""
        pass
    
    def iterate(self):
        """
        Called to perform one iteration of the event loop.
        
        You *shouldn't* have to override the core functionality, but you may
        wish to override this method to do something before or after.
        """
        # Constant FPS, Potter!
        self.clock.tick(self.ticks)
        # Change to the new state (if there is one)
        if self.state != self.nextState:
            self.state = self.nextState
            self.state.firstDisplay(self.screen)
        # Send events to the state
        for event in pygame.event.get():
            if event.type == sound.music.endEvent:
                sound.music.nextSong()
            else:
                self.state.handle(event)
        # Update state
        self.state.update(self)
        # Display state
        self.state.display(self.screen)
    
    def postloop(self):
        """Hook method to do something after the event loop exits."""
        pass

class FakeClock(object):
    """
    Imitator of ``pygame.time.Clock`` that doesn't delay.
    
    Used for `TwistedGame`.
    """
    def __init__(self):
        self.time = pygame.time.get_ticks()
        self.delta = 0
        self.delay = 0
    
    def get_fps(self):
        """
        Returns the current rate of frames per second.
        
        :ReturnType: float
        """
        return 1 / self.get_time()
    
    def get_rawtime(self):
        """
        Returns number of nondelayed milliseconds between last two calls to
        `tick`.
        
        :ReturnType: int
        """
        return self.delta
    
    def get_delay(self):
        """
        Returns number of delayed milliseconds between last two calls to `tick`.
        
        :ReturnType: int
        """
        return self.delay
    
    def get_time(self):
        """
        Returns number of milliseconds between last two calls to `tick`.
        
        :ReturnType: int
        """
        return self.get_rawtime() + self.get_delay()
    
    def tick(self, ticks=None):
        """
        Keep running at a given framerate.
        
        :Returns: The time since the last call to `tick` in milliseconds.
        :ReturnType: int
        """
        newTime = pygame.time.get_ticks()
        self.delta = newTime - self.time
        self.time = newTime
        if ticks:
            assert ticks > 0
            self.delay = int(max((1 / ticks) * 1000 - self.delta, 0))
        else:
            self.delay = 0
        return self.delta

class TwistedGame(Game):
    """
    Game object that uses a Twisted_ reactor for the main event loop.
    
    .. _Twisted: http://twistedmatrix.com/
    """
    def __init__(self, root_dir, reactor=None, **kw):
        """
        Initialize game object.
        
        :Parameters:
            root_dir : string
                Directory to chdir to.  You can even pass sys.argv[0].
            reactor : Twisted reactor
                Reactor to use for main event loop.  If not specified, then the
                default reactor is used.
        """
        super(TwistedGame, self).__init__(root_dir, **kw)
        if reactor is None:
            from twisted.internet import reactor
        self.reactor = reactor
    
    def run(self):
        """
        Run the game.
        
        This sets up a `FakeClock` object and starts the reactor.
        """
        # Do usual initialization
        pygame.init()
        self.screen = pygame.display.set_mode(self.screenSize, self.flags)
        self.clock = FakeClock()
        self.preloop()
        self.running = True
        # Iterate will tell the reactor to keep calling it
        self.iterate()
        self.reactor.run()
        # Do post loop
        self.postloop()
    
    def end(self):
        """Stops the game and reactor."""
        super(TwistedGame, self).end()
        self.reactor.stop()
    
    def iterate(self):
        """
        Runs an iteration of the event loop, then tells the reactor to iterate
        again.
        """
        try:
            super(TwistedGame, self).iterate()
            self.reactor.callLater(self.clock.get_delay() / 1000, self.iterate)
        except SystemExit:
            self.end()



#!/usr/bin/env python
#
#   timer.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""Timing utilities"""

from __future__ import division

__author__ = 'Ross Light'
__date__ = 'January 5, 2007'
__all__ = ['Timer',
           'TimerGroup',]
__docformat__ = 'reStructuredText'

class Timer(object):
    """
    Timer that calls a callback after a given amout of time.
    
    :IVariables:
        duration : float
            Length of time (in seconds) between timer firings
        loop : int
            The number of times to loop.  If zero (the default), the timer fires
            only once.  If its a positive integer, it fires that many times plus
            one.  Otherwise, the timer fires indefinitely.
        callback : function
            Function to call when timer fires
        callArgs : tuple
            Additional arguments to pass to callback
        callKw : tuple
            Additional keyword arguments to pass to callback
    """
    def __init__(self,
                 duration,
                 loop=0,
                 callback=None,
                 callArgs=None,
                 callKw=None):
        """
        Initializes the timer.
        
        :Parameters:
            duration : float
                Length of time between timer firings
            loop : int
                The number of times to loop
            callback : function
                Function to call when timer fires
            callArgs : tuple
                Additional arguments to pass to callback
            callKw : tuple
                Additional keyword arguments to pass to callback
        """
        self.callback = callback
        if callArgs is None:
            self.callArgs = []
        else:
            self.callArgs = list(callArgs)
        if callKw is None:
            self.callKw = {}
        else:
            self.callKw = dict(callKw)
        self.duration = float(duration)
        self.loop = loop
        self.reset()
    
    def reset(self, duration=None):
        """
        Resets the timer.
        
        :Parameters:
            duration : float
                The timer's new duration (if given)
        """
        if duration is not None:
            self.duration = float(duration)
        if self.duration < 0:
            raise ValueError("Duration must be greater than 0")
        self.value = self.duration
    
    def update(self, time):
        """
        Updates the timer count.
        
        :Parameters:
            time : float
                The elapsed time.  This can be a number, in which case it is
                assumed to be the elapsed time in seconds; an object with a
                ``get_time`` method (such as a ``pygame.time.Clock``
                instance); or an object with a ``clock`` attribute, which should
                provide a ``get_time`` method.  The ``get_time`` should return
                the elapsed time in milliseconds.
        :Returns: How many times the timer fired
        :ReturnType: int
        """
        # If we're done, do nothing
        if self.loop == 0 and self.value == 0:
            return 0
        # Get the time elapsed
        if hasattr(time, 'get_time'):
            time = time.get_time() / 1000
        elif hasattr(time, 'clock'):
            time = time.clock.get_time() / 1000
        else:
            try:
                time = float(time)
            except TypeError:
                raise TypeError("Unknown type given to Timer.update()")
        # Fire as many times as necessary
        fireCount = 0
        self.value -= time
        while self.value <= 0:
            fireCount += 1
            self.value += self.duration
            if self.callback is not None:
                self.callback(*self.callArgs, **self.callKw)
            if self.loop == 0:
                self.value = 0.0
                break
            elif self.loop > 0:
                self.loop -= 1
        return fireCount

class TimerGroup(object):
    """
    A group of `Timer` objects that all update with one call.
    
    :IVariables:
        timers : list
            List of timers in group
    """
    def __init__(self, timers=None):
        """
        Initializes the group.
    
        :Parameters:
            timers : list
                List of timers in group
        """
        if timers is None:
            timers = []
        self.timers = list(timers)
    
    def add(self, timer):
        """
        Adds a timer to the group.
    
        :Parameters:
            timer : `Timer`
                Timer to add.
        """
        self.timers.append(timer)
    
    def create(self, *args, **kw):
        """
        Creates a new timer and adds it to the group.
        
        :See: `Timer.__init__`
        """
        self.add(Timer(*args, **kw))
    
    def remove(self, timer):
        """
        Removes a timer from the group.
    
        :Parameters:
            timer : `Timer`
                Timer to remove.
        """
        self.timers.remove(timer)
    
    def update(self, time):
        """
        Updates all of the timers.
    
        :See: `Timer.update`
        """
        for timer in self:
            timer.update(time)
    
    def __len__(self):
        return len(self.timers)
    
    def __iter__(self):
        return iter(self.timers)



#!/usr/bin/env pythonw
#
#   sound.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""
pymage sound subsystem

:Variables:
    music : `MusicManager`
        The global music manager
    sound : `SoundManager`
        The global sound manager
"""

import warnings

import pygame
from pygame.locals import *

from pymage import resman

__author__ = 'Ross Light'
__date__ = 'July 19, 2006'
__all__ = ['MusicManager',
           'music',
           'SoundManager',
           'sound',]
__docformat__ = 'reStructuredText'

class MusicManager(resman.Submanager):
    """
    Game music manager.
    
    For sound effects, use `SoundManager`.
    
    :CVariables:
        endEvent : int
            The event indicating the end of a song
    :IVariables:
        savedPlaylists : dict of lists
            Saved playlists
        playlist : list
            Current play order
        index : int
            Index of the current song in the current playlist
        loop : bool
            Whether the music should loop
        shouldPlay : bool
            Whether music should be loaded
        playing : bool
            Whether music is playing
        volume : float
            Volume of music (from 0.0 to 1.0)
    """
    endEvent = 25
    
    def __init__(self,
                 should_play=True,
                 loop=True,
                 *args, **kw):
        """
        Initializes the music manager.
        
        :Parameters:
            should_play : bool
                Whether music should be loaded
            loop : bool
                Whether to loop music
            manager : `ResourceManager`
                The resource manager to use.  Default is `pymage.resman.resman`.
        """
        super(MusicManager, self).__init__(*args, **kw)
        self.savedPlaylists = {}
        self.playlist = []
        self.loop = loop
        self.shouldPlay = should_play
        self.index = 0
        self.playing = False
        pygame.mixer.music.set_endevent(self.endEvent)
    
    # Playlist management
    
    def prepare(self, tag, playlist):
        """
        Prepares a playlist for playback later.
        
        .. Warning::
           `prepare` is retained for compatibility reasons.  You should be using
           `addPlaylist` to add new playlists.
        
        :Parameters:
            tag : string
                Name of new playlist
            playlist : list of strings
                Songs in playlist
        """
        warnings.warn("prepare is deprecated; use addPlaylist.",
                      DeprecationWarning,
                      stacklevel=2)
        self.addPlaylist(tag, playlist)
    
    def addPlaylist(self, key, songs):
        """
        Adds a new playlist.
        
        :Parameters:
            key : string
                Name of new playlist
            songs : list of strings
                Songs in new playlist
        """
        self.savedPlaylists[key] = list(songs)
    
    def getPlaylist(self, key):
        """
        Accesses the playlist.
        
        :Parameters:
            key : string
                Name of playlist
        :Returns: List of songs in playlist
        :ReturnType: list
        :Raises KeyError: If playlist does not exist
        """
        return self.savedPlaylists[key]
    
    def removePlaylist(self, key):
        """
        Removes a saved playlist.
        
        :Parameters:
            key : string
                Name of playlist
        :Raises KeyError: If playlist does not exist
        """
        del self.savedPlaylists[key]
    
    # Playback control
    
    def startPlaylist(self, key):
        """
        Starts a playlist denoted by ``key``.
        
        After calling `startPlaylist`, the playlist is *cued*.  You then need to
        call `play` to start playing music.
        
        :Parameters:
            key : string or iterable
                Playlist to play.  If an iterable is passed, then it will be
                used as a temporary playlist.
        """
        try:
            self.playlist = self.getPlaylist(key)
        except KeyError:
            try:
                self.playlist = list(key)
            except TypeError:
                raise TypeError("Playlist must be added or an iterable object")
        if self.playlist:
            self.index = 0
            self.loadSong()
    
    def play(self):
        """Starts the current song."""
        if self.shouldPlay and not self.playing:
            if pygame.mixer.music.get_busy():
                pygame.mixer.music.unpause()
            else:
                pygame.mixer.music.play()
            self.playing = True
    
    def pause(self):
        """
        Pauses the current song.
        
        You can resume using `play`.
        """
        pygame.mixer.music.pause()
        self.playing = False
    
    def stop(self):
        """Stops the current song and goes to the beginning."""
        pygame.mixer.music.stop()
        self.playing = False
    
    def load(self, key=None):
        """
        Manually loads a song.
        
        This method will not load anything unless the manager is configured to
        play.
        
        :Parameters:
            key : string
                Name of song.  If not given, the current song is loaded.
        """
        if key is None:
            if self.playlist:
                key = self.playlist[self.index]
            else:
                raise ValueError("No song in the playlist")
        if self.shouldPlay:
            super(MusicManager, self).load(key)
    
    loadSong = load
    
    def previousSong(self):
        """Returns to the previous song in the playlist."""
        if self.playlist:
            self.index -= 1
            if self.index < 0:
                self.index = len(self.playlist) - 1
            self.stop()
            self.loadSong()
        
    def nextSong(self):
        """Advances to the next song in the playlist."""
        if self.playlist:
            playNextSong = self.playing
            self.index += 1
            if self.index >= len(self.playlist):
                self.index = 0
                if not self.loop:
                    playNextSong = False
            self.stop()
            self.loadSong()
            if playNextSong:
                self.play()
    
    # Volume control
    
    def getVolume(self):
        """
        Gets the current volume.
        
        You can also use the `volume` property.
        """
        return self.volume
    
    def setVolume(self, volume):
        """
        Changes the current volume.
        
        You can also use the `volume` property.
        """
        self.volume = volume
    
    def _getVolume(self):
        """Retrieves the mixer's current volume."""
        return pygame.mixer.music.get_volume()
    
    def _setVolume(self, volume):
        """Changes the mixer's current volume."""
        pygame.mixer.music.set_volume(volume)
    
    volume = property(_getVolume, _setVolume, doc="The mixer's volume")

music = MusicManager()

class SoundManager(resman.Submanager):
    """
    Sound effects manager.
    
    For music, use `MusicManager`.
    
    :IVariables:
        shouldPlay : bool
            Whether sound effects should be loaded
        volume : float
            Volume of played sound effects (from 0.0 to 1.0)
    """
    resourceType = resman.SoundResource
    
    def __init__(self, should_play=True, volume=0.5, *args, **kw):
        """
        Initializes the sound manager.
        
        :Parameters:
            should_play : bool
                Whether sound effects should be loaded
            volume : float
                Volume of played sound effects (from 0.0 to 1.0)
            manager : `ResourceManager`
                The resource manager to use.  Default is `pymage.resman.resman`.
        """
        super(SoundManager, self).__init__(*args, **kw)
        self.shouldPlay = should_play
        self.volume = volume
    
    def getSound(self, *args, **kw):
        """
        Retrieves a sound, using a cache if possible.

        .. Warning::
           `getSound` is deprecated, for favor of the Submanager API.  Use
           `load` instead.
        
        :Parameters:
            key : string
                Name of sound
        :Returns: The sound requested
        :ReturnType: ``pymage.mixer.Sound``
        """
        warnings.warn("getSound is deprecated; use load.",
                      DeprecationWarning,
                      stacklevel=2)
        return self.load(*args, **kw)
    
    def play(self, tag, volume=None, cache=True):
        """
        Plays a sound.
        
        :Parameters:
            tag : string
                Name of sound effect
            volume : float
                Volume of the sound effect.  If not specified, `volume`
                attribute is used
            cache : bool
                Whether the sound will be cached
        :Returns: The playing sound
        :ReturnType: ``pymage.mixer.Sound``
        """
        if volume is None:
            volume = self.volume
        if self.shouldPlay:
            if cache:
                self.cache(tag)
            snd = self.load(tag)
            snd.set_volume(volume)
            snd.play()
            return snd
        else:
            return None

sound = SoundManager()



#!/usr/bin/env python
#
#   setup.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""setuptools script for pymage"""

# Run bootstrap setuptools installer
from ez_setup import use_setuptools
use_setuptools()

# Import modules
from setuptools import setup
import sys

__author__ = 'Ross Light'
__date__ = 'August 16, 2006'
__all__ = []

# Check for Python 2.4
if sys.version < '2.4.0':
    print >> sys.stderr, "pymage requires Python 2.4"
    sys.exit(1)

# Call setup function
setup(name="pymage",
      version='0.2.2',
      description="Pygame helper package",
      long_description="""\
pymage is a Python package that simplifies many aspects of Pygame programming
(e.g. resource loading).

Its features include:

* File-based configuration
* Auto-configuring joystick support
* Music playlists
* Vectors
* State machine
* High-level timing""",
      license='LGPL',
      author="Ross Light",
      author_email='rlight2@gmail.com',
      keywords="pygame SDL configuration resource",
      url='http://code.google.com/p/pymage/',
      download_url='http://code.google.com/p/pymage/downloads/list',
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Games/Entertainment',],
      test_suite='tests',
      packages=['pymage'],
      install_requires=['pygame'],
      extras_require={
        'OpenGL': ["PyOpenGL"],
        'Twisted': ["Twisted"],
      },
      zip_safe=True,)



#!/usr/bin/env python
#
#   config.py
#   pymage
#
#   Copyright (C) 2006-2007 Ross Light
#
#   This file is part of pymage.
#
#   pymage is free software; you can redistribute it and/or modify it under the
#   terms of the GNU Lesser General Public License as published by the Free
#   Software Foundation; either version 3 of the License, or (at your option)
#   any later version.
#   
#   pymage is distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
#   more details.
#   
#   You should have received a copy of the GNU Lesser General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Provide configuration for your game.

The typical use of this module is to use the `setup` function and give set up
parameters.  From there, the module sets up the correct parts of the pymage
library.

This module is not strictly necessary, but pymage's reasoning is as follows:

Resources for the game should be loaded on an location-independent basis.
This provides abstraction and the ability for arbitrary file structure.
However, these files eventually need to be located, so the developer should
create a file (separate from the main code) that locates everything.  From
there, there may be a legitimate reason for the user to want to customize the
file locations (perhaps he/she wants the game resources to be on a separate hard
drive).  Therefore, a separate file is used to override the behavior of any
given installation of the game.  But since the user never touches the
developer's default game setup, any corrupt preferences can be corrected by
deleting the user configuration files.

Here is a sample gamesite.xml file::

    <?xml version="1.0"?>
    
    <!-- name is not necessary, but reserved for future use. -->
    <game-site name="Sample Game">
        <!-- Prepare an image in the ImageManager.
             The id attribute is used as the tag.
             The path element is required to find the file.
             The section and option elements specify how the path can be
             overriden in configuration files. -->
        <image id="SampleImage">
            <section>sprites</section>
            <option>sample</option>
            <path>img/sample.png</path>
        </image>
        
        <!-- Prepare a sound in the SoundManager.
             The id attribute is used as the tag.
             The path element is required to find the file.
             The section and option elements specify how the path can be
             overriden in configuration files. -->
        <sound id="SampleSound">
            <section>sounds</section>
            <option>sample</option>
            <path>sfx/sample.wav</path>
        </sound>
        
        <!-- Prepare a playlist in the MusicManager.
             The id attribute is used as the tag.
             The path element(s) are required to find the music file(s).
             The section and option elements specify how the playlist can be
             overriden in configuration files. -->
        <music id="song1">
            <path>music/song1.wav</path>
        </music>
        <playlist id="SamplePlaylist">
            <section>playlists</section>
            <option>sample</option>
            <music ref="song1"/>
            <path>music/song2.wav</path> <!-- Old way, still works -->
        </playlist>
        
        <!-- Prepare a cache group. -->
        <group id="SampleGroup">
            <section>groups</section>
            <option>sample</option>
            <image ref="SampleImage"/>
            <sound ref="SampleSound"/>
        </group>
        
        <!-- Specify additional configuration files. -->
        <config-file>userconfig.ini</config-file>
    </game-site>
"""

from ConfigParser import ConfigParser
import os
from textwrap import dedent
import warnings
from xml.dom import minidom

from pymage import resman
from pymage import sound

__author__ = 'Ross Light'
__date__ = 'August 10, 2006'
__all__ = ['GameSiteWarning',
           'load',
           'save',
           'getOption',
           'setOption',
           'registerType',
           'unregisterType',
           'setup',]
__docformat__ = 'reStructuredText'

# Globals
_gsPrims = {'resource': resman.Resource,
            'image': resman.ImageResource,
            'sound': resman.SoundResource,
            'music': resman.MusicResource,}

class GameSiteWarning(UserWarning):
    """Warning emitted when odd game site constructs are used."""
    pass

## CONFIG PARSER ##

class CaseConfigParser(ConfigParser):
    """A ``ConfigParser`` that is case-sensitive."""
    def optionxform(self, optstr):
        """Return the string in the same case."""
        return optstr

def load(*args, **kw):
    """
    Load a series of configuration files.
    
    Any keyword values are used as variables.
    
    :Keywords:
        convert : bool
            Whether the function should interpret the values as what they seem
            to be (e.g. ``float``, ``int``).  The default is ``True``.
    :Returns: Loaded configuration
    :ReturnType: dict
    """
    # Retrieve convert keyword
    convertValues = kw.get('convert', True)
    try:
        del kw['convert']
    except KeyError:
        pass
    # Parse the files
    parser = CaseConfigParser(kw)
    for configFile in args:
        close = False
        if isinstance(configFile, basestring):
            # Open strings as paths
            path = os.path.normpath(os.path.expanduser(configFile))
            if os.path.exists(path):
                configFile = open(path)
            else:
                continue
            close = True
        parser.readfp(configFile)
        if close:
            configFile.close()
    # Assemble dictionary
    configDict = {}
    for section in parser.sections():
        sectionDict = {}
        for option in parser.options(section):
            value = parser.get(section, option)
            if convertValues:   # Interpret values
                value = _getValue(value)
            sectionDict[option] = value
        configDict[section] = sectionDict
    return configDict

def save(config, configFile):
    """
    Saves a configuration dictionary to a file.
    
    :Parameters:
        config : dict
            Configuration dictionary
        configFile : string or file
            File to write configuration to
    """
    close = False
    parser = CaseConfigParser()
    for section, values in config.iteritems():
        for option, value in values.iteritems():
            parser.set(section, option, value)
    if isinstance(configFile, basestring):
        path = os.path.normpath(os.path.expanduser(configFile))
        configFile = open(path, 'w')
        close = True
    parser.write(configFile)
    if close:
        configFile.close()

def _getValue(value_string):
    """
    Retrieves a value from a ``ConfigParser`` string.
    
    :Parameters:
        value_string : string
            Option string to convert
    :Returns: The string's value, converted into an int, bool, float, or string
    """
    boolLiterals = {'false': False,
                    'no': False,
                    'off': False,
                    'true': True,
                    'yes': True,
                    'on': True,}
    if value_string.isdigit():
        # Integer
        return int(value_string)
    elif value_string.lower() in boolLiterals:
        # Boolean
        return boolLiterals[value_string.lower()]
    elif _isFloat(value_string):
        # Float
        return float(value_string)
    else:
        # String
        return str(value_string)

def _isFloat(value_string):
    """
    Returns whether the string is a ``float``.
    
    The format for a float is::
    
        int[.[fraction]]
    
    :Parameters:
        value_string : string
            String to test for floatiness
    :ReturnType: bool
    """
    if value_string.isdigit():
        # int
        return True
    elif value_string.count('.') == 1:
        # has a decimal point
        value_string = value_string.replace('.', '')
        if value_string.isdigit():
            # int[.[fraction]]
            return True
        else:
            return False
    else:
        return False

def getOption(config, section, option, default=None):
    """
    Retrieves an option from the configuration dictionary.
    
    :Parameters:
        config : dict
            Configuration dictionary to read from
        section : string
            Configuration section
        option : string
            Option name
        default
            Default value to return if option is not found
    :Returns: The requested option's value
    """
    try:
        section = config[section]
    except KeyError:
        return default
    else:
        return section.get(option, default)

def setOption(config, section, option, value):
    """
    Changes an option in the configuration dictionary.
    
    :Parameters:
        config : dict
            Configuration dictionary to modify
        section : string
            Configuration section
        option : string
            Option name
        value
            Value to change the option to
    """
    section = config.setdefault(section, {})
    section[option] = value

## GAME SITE ##

def registerType(tag, factory):
    """
    Register a custom game site resource type.
    
    :Parameters:
        tag : string
            Name of the XML element
        factory
            Callable that takes one positional argument: the path to the
            resource.  Any additional attributes found on the XML element are
            passed as keyword arguments.  The value of this parameter is
            typically the constructor of a `pymage.resman.Resource` subclass.
    """
    _gsPrims[tag] = factory

def unregisterType(tag):
    """
    Unregister a custom game site resource type.
    
    :Parameters:
        tag : string
            Name of the XML element
    """
    del _gsPrims[tag]

def setup(site='gamesite.xml', *config_files, **kw):
    """
    Sets up a game from the specified parameters and returns the configuration
    dictionary.
    
    The additional arguments are program default configuration files.  These
    are parsed before any inside the game site file (therefore giving the site
    configuration files higher precedence).
    
    :Parameters:
        site : string or file
            Game site file.
    :Keywords:
        configSound : bool
            Whether the sound manager volume should be configured automatically
        configMusic : bool
            Whether the music manager volume should be configured automatically
    """
    # Get keyword arguments
    configSound = kw.pop('configSound', True)
    configMusic = kw.pop('configMusic', True)
    if kw:
        raise TypeError("Invalid keyword argument")
    # Parse game site file
    doc = minidom.parse(site)
    config = _getSiteConfig(doc, config_files)
    # Load configuration
    if configSound:
        _processSoundOptions(config)
    if configMusic:
        _processMusicOptions(config)
    # Process resources
    _processGameSite(doc, config)
    # Return configuration dictionary
    return config

def _getSiteConfig(doc, config_files):
    """
    Obtains full configuration.
    
    The configuration files passed in are put first in the list, so the ones
    specified in the game site file take precedence.
    
    :Parameters:
        doc : DOM document
            Game site DOM tree
        config_files : list of strings or files
            Configuration files to load
    :Returns: The loaded configuration dictionary
    :ReturnType: dict
    """
    siteConfigs = []
    for child in doc.documentElement.childNodes:
        if (child.nodeType == minidom.Node.ELEMENT_NODE and
            child.tagName == 'config-file'):
            siteConfigs.append(_getText(child))
    config_files = list(config_files) + list(siteConfigs)
    return load(*config_files)

def _processSoundOptions(config):
    """
    Configure the sound manager with the default configuration keys.
    
    :Parameters:
        config : dict
            The configuration dictionary
    """
    sound.sound.shouldPlay = bool(getOption(config, 'sound', 'play', True))
    sound.sound.volume = float(getOption(config, 'sound', 'volume', 1.0))

def _processMusicOptions(config):
    """
    Configure the music manager with the default configuration keys.
    
    :Parameters:
        config : dict
            The configuration dictionary
    """
    sound.music.shouldPlay = bool(getOption(config, 'music', 'play', True))
    sound.music.volume = bool(getOption(config, 'music', 'volume', 0.5))
    sound.music.loop = bool(getOption(config, 'music', 'loop', True))

def _processGameSite(doc, config):
    """
    Run through game site file and add resources to manager.
    
    :Parameters:
        doc : DOM document
            Game site DOM tree
        config : dict
            Configuration dictionary
    """
    handlers = {'playlist': _handlePlaylist,
                'group': _handleGroup,}
    handlers.update(dict.fromkeys(_gsPrims, _handlePrimitive))
    for child in doc.documentElement.childNodes:
        if (child.nodeType == minidom.Node.ELEMENT_NODE and
            child.tagName in handlers):
            # Call handler
            handler = handlers[child.tagName]
            handler(child, config)

def _handlePrimitive(elem, config):
    """
    Handle a basic resource (i.e. images, sound effects, and custom resources).
    
    :Parameters:
        elem : DOM node
            Element to handle
        config : dict
            Configuration dictionary
    :Returns: Resource's key
    :ReturnType: string
    """
    attr = _attributes(elem, include_ns=False, ascii=True)
    pathChild = _childNamed(elem, 'path')
    if pathChild is None:
        warnings.warn("Primitive without a path", GameSiteWarning)
        return
    # Get ID
    try:
        key = attr.pop('id')
    except KeyError:
        pathChild = _childNamed(elem, 'path')
        if pathChild:
            key = _getText(pathChild)
        else:
            warnings.warn("Primitive without a key", GameSiteWarning)
            return
    # Get resource information
    resType = _gsPrims[elem.tagName]
    section = _getText(_childNamed(elem, 'section'))
    option = _getText(_childNamed(elem, 'option'))
    path = _getText(pathChild)
    # Create resource
    if section is not None and option is not None:
        path = getOption(config, section, option, path)
    resman.resman.addResource(key, resType(path, **attr))
    # Return key
    return key

def _handlePlaylist(elem, config):
    """
    Handle a playlist element.
    
    :Parameters:
        elem : DOM node
            Element to handle
        config : dict
            Configuration dictionary
    :Returns: Playlist's key
    :ReturnType: string
    """
    key = elem.getAttribute('id')
    section = _getText(_childNamed(elem, 'section'))
    option = _getText(_childNamed(elem, 'option'))
    playlistKeys = []
    # Get playlist keys
    for sub in elem.childNodes:
        if sub.nodeType == minidom.Node.ELEMENT_NODE:
            if sub.tagName == 'path':
                # Old-school path approach
                warnings.warn("%s using old path-based playlist" % (key),
                              GameSiteWarning)
                if sub.hasAttribute('id'):
                    musicKey = sub.getAttribute('id')
                    musicPath = _getText(sub)
                else:
                    musicKey = musicPath = _getText(sub)
                resman.addResource(musicKey, resman.MusicResource(musicPath))
                playlistKeys.append(musicKey)
            elif sub.tagName == 'music':
                # New-school music reference/declaration approach
                if sub.hasAttribute('ref'):
                    musicKey = sub.getAttribute('ref')
                else:
                    musicKey = _handlePrimitive(sub, config)
                playlistKeys.append(musicKey)
    # Create playlist
    if section is not None and option is not None:
        configKeys = getOption(config, section, option)
        if configKeys is not None:
            playlistKeys = configKeys.split(',')
    sound.music.addPlaylist(key, playlistKeys)
    # Return key
    return key

def _handleGroup(elem, config):
    """
    Handle a group element (a cache group).
    
    :Parameters:
        elem : DOM node
            Element to handle
        config : dict
            Configuration dictionary
    :Returns: Cache group's key
    :ReturnType: string
    """
    key = elem.getAttribute('id')
    section = _getText(_childNamed(elem, 'section'))
    option = _getText(_childNamed(elem, 'option'))
    groupKeys = set()
    # Get group keys
    for sub in elem.childNodes:
        if (sub.nodeType == minidom.Node.ELEMENT_NODE and
            sub.tagName in _gsPrims):
            if sub.hasAttribute('ref'):
                resourceKey = sub.getAttribute('ref')
            else:
                resourceKey = _handlePrimitive(sub, config)
            groupKeys.add(resourceKey)
    # Create group
    if section is not None and option is not None:
        configKeys = getOption(config, section, option)
        if configKeys is not None:
            groupKeys = configKeys.split(',')
    resman.resman.addCacheGroup(key, groupKeys)
    # Return key
    return key

def _getText(elem, post=True):
    """
    Retrieve text from a DOM node, stripping indents, if asked.
    
    This function does honor the ``xml:space`` attribute, and if
    ``xml:space="preserve"`` is specified, it takes precendence over the
    ``post`` argument.
    
    :Parameters:
        elem : DOM node
            The element to get text from
    :Keywords:
        post : bool
            Whether to strip indents
    :Returns: The element's text
    :ReturnType: string
    """
    xmlNS = 'http://www.w3.org/XML/1998/namespace'
    if elem is None:
        return None
    text = ''
    for child in elem.childNodes:
        if child.nodeType == minidom.Node.TEXT_NODE:
            text += child.wholeText
    preserve = (elem.hasAttributeNS(xmlNS, 'space') and
                elem.getAttributeNS(xmlNS, 'space') == 'preserve')
    if post and not preserve:
        text = dedent(text)
        if text.startswith('\n'):
            text = text[1:]
            if text.endswith('\n'):
                text = text[:-1]
    return text

def _childNamed(elem, name):
    """
    Returns the first child with the given name.
    
    :Parameters:
        elem : DOM node
            The element to search in
        name : string
            The name to search for
    :Returns: The named child, or ``None`` if not found
    :ReturnType: DOM node
    """
    for child in elem.childNodes:
        if (child.nodeType == minidom.Node.ELEMENT_NODE and
            child.tagName == name):
            return child
    else:
        return None

def _attributes(elem, include_ns=True, ascii=False):
    """
    Retrieves the attributes of a DOM node as a dictionary.
    
    If ``include_ns`` is ``True``, then .  If ``ascii`` is ``True``, then 
    
    :Parameters:
        elem : DOM node
            The element to extract attributes from
    :Keywords:
        include_ns : bool
            Whether attributes with a namespace will be discarded
        ascii : bool
            Whether the attribute names are converted to ASCII.  If an attribute
            name cannot be converted, the entire attribute is discarded.
    """
    nodemap = elem.attributes
    attrDict = {}
    for index in xrange(nodemap.length):
        attr = nodemap.item(index)
        if not include_ns and attr.prefix:
            continue
        name = attr.localName
        if ascii:
            try:
                name = str(name)
            except UnicodeError:
                continue
        attrDict[name] = attr.value
    return attrDict
